<?php

namespace App\Http\Controllers;
require __DIR__.'../../../../vendor/autoload.php';

use App\Helper\Helper;
use App\Models\Location;
use App\Models\UserActivity;
use App\Models\UserLocation;
use Illuminate\Http\JsonResponse;
use Illuminate\Http\Request;
use App\Models\User;
use Exception;
use Illuminate\Support\Facades\DB;
use MailchimpTransactional\ApiClient;
use GuzzleHttp\Exception\ServerException;

class UserController extends Controller
{
    //

    public function signUp(Request $request)
    {

        $request->validate([
            'profile_picture' => 'required',
            'type' => 'required',
            'name' => 'required',
            'email' => 'required | unique:users,email',
            'password' => 'required | confirmed'
        ]);

        try {

            DB::beginTransaction();

            $user = User::create([
                'name' => $request->name,
                'email' => $request->email,
                'password' => bcrypt($request->password),
            ]);
            $user->plan_id = Helper::FREE_PLAN;
            $user->plan_id_tourist = Helper::FREE_PLAN;

            switch ($request->type) {
                case Helper::TOURIST_TYPE:
                    $user->assignRole(Helper::TOURIST_TYPE);
                    break;
                case Helper::LOCAL_TYPE:
                    $user->assignRole(Helper::LOCAL_TYPE);
                    break;
                default:
                    # code...
                    DB::rollBack();
                    return response()->json([
                        'message' => 'Invalid user type selected. It can be either local or tourist.',
                    ], 422);
                    break;
            }

            // add profile picture
            $file = $request->file('profile_picture');
            $destinationPath = 'uploads/user_profile_pictures/'.$user->id;
            $file_uploaded = $file->move($destinationPath, $file->getClientOriginalName());

            $full_unique_path = $file_uploaded->getPath() . '/'. $file_uploaded->getFilename();

            $user->profile_picture = asset($full_unique_path);
            $user->save();

            DB::commit();

            $this->sendRegistrationEmail($user);

            return response()->json([
                'user' => $user,
                'token' => $user->createToken('mobile_app')->plainTextToken,
            ], 200);
        } catch (\Exception $e) {
            return response()->json([
                'message' => $e->getMessage(),
            ], 500);
        }
    }

    /**
     * @param Request $request
     * @return JsonResponse
     */
    public function locals(Request $request): JsonResponse
    {
        return response()->json([
            'locals' => User::role(Helper::LOCAL_TYPE)->with('locations', 'locations.location', 'activities', 'activities.activity')->get(),
        ]);
    }

    /**
     * @param Request $request
     * @return JsonResponse
     */
    public function tourists(Request $request): JsonResponse
    {
        return response()->json([
            'tourists' => User::role(Helper::TOURIST_TYPE)->get(),
        ]);
    }

    /**
     * @param Request $request
     * @return JsonResponse
     */
    public function filter_locals(Request $request): JsonResponse
    {
        $locals = User::role(Helper::LOCAL_TYPE);

        if ($request->min_range && $request->max_range){
            $locals->whereBetween('charges_per_hour', [(int)$request->min_range, (int)$request->max_range]);
        }

        if ($request->language){
            $locals->where('language', $request->language);
        }

        $locals = $locals->get();

        if ($request->has('activities')){
            if (!empty($request->input('activities'))){
                foreach ($locals as $key => $local) {
                    $local_activites = UserActivity::where('user_id', $local->id)->get();
                    $requested_activities_exist_in_local = false;
                    foreach ($request->activities as $request_activity_id) {
                        foreach ($local_activites as $local_activite){
                            if ($local_activite->activity_id == $request_activity_id){
                                $requested_activities_exist_in_local = true;
                            }
                        }
                    }
                    if (!$requested_activities_exist_in_local){
                        $locals->forget($key);
                    }
                }
            }
        }

        return response()->json([
            'locals' => $locals,
        ]);
    }

    /**
     * @param Request $request
     * @return JsonResponse
     */
    public function search_locals(Request $request): JsonResponse
    {
        $request->validate([
            'search_string' => 'required'
        ]);

        return response()->json([
            'locals' => User::role(Helper::LOCAL_TYPE)->where('name', 'LIKE', '%'.$request->search_string.'%')->get(),
        ]);
    }

    /**
     * @param Request $request
     * @return JsonResponse
     */
    public function activities_get_all(Request $request): JsonResponse
    {
        return response()->json([
            'activities' => Helper::LOCAL_ACTIVITES,
        ]);
    }

    /**
     * @param Request $request
     * @return JsonResponse
     */
    public function locations_get_all(Request $request): JsonResponse
    {
        return response()->json([
            'locations' => Location::all(),
        ]);
    }

    /**
     * @param Request $request
     * @return JsonResponse
     */
    public function update_user_general_settings(Request $request): JsonResponse
    {

        try {
            $user = $request->user();

            if ($request->bg_image){
                // add profile picture
                $file = $request->file('bg_image');
                $destinationPath = 'uploads/user_bg_images/'.$user->id;
                $file_uploaded = $file->move($destinationPath, $file->getClientOriginalName());

                $full_unique_path = $file_uploaded->getPath() . '/'. $file_uploaded->getFilename();

                $user->bg_image = asset($full_unique_path);
                $user->save();
            }

            if ($request->profile_picture){
                // add profile picture
                $file = $request->file('profile_picture');
                $destinationPath = 'uploads/user_profile_pictures/'.$user->id;
                $file_uploaded = $file->move($destinationPath, $file->getClientOriginalName());

                $full_unique_path = $file_uploaded->getPath() . '/'. $file_uploaded->getFilename();

                $user->profile_picture = asset($full_unique_path);
                $user->save();
            }

            if ($request->name){
                $user->name = $request->name;
                $user->save();
            }

            if ($request->about){
                $user->about = $request->about;
                $user->save();
            }

            if ($request->charges_per_hour){
                $user->charges_per_hour = $request->charges_per_hour;
                $user->save();
            }

            if ($request->currency){
                $user->currency = $request->currency;
                $user->save();
            }

            if ($request->language){
                $user->language = $request->language;
                $user->save();
            }

            if ($request->gender){
                $user->gender = $request->gender;
                $user->save();
            }

            if ($request->locations && !empty($request->locations)){
                $user_locations = UserLocation::where('user_id', $user->id)->get();
                foreach ($user_locations as $user_location){
                    $user_location->delete();
                }
                foreach ($request->locations as $location_id){
                    $location = Location::find($location_id);
                    if ($location){
                        UserLocation::create(
                            [
                                'user_id' => $user->id,
                                'location_id' => $location->id
                            ]
                        );
                    }
                }
            }

            if ($request->activities && !empty($request->activities)){
                $user_activities = UserActivity::where('user_id', $user->id)->get();
                foreach ($user_activities as $user_activity){
                    $user_activity->delete();
                }
                foreach ($request->activities as $activity_id){
                    $activity = array_key_exists($activity_id, Helper::LOCAL_ACTIVITES);
                    if ($activity){
                        UserActivity::create(
                            [
                                'user_id' => $user->id,
                                'activity_id' => $activity_id
                            ]
                        );
                    }
                }
            }

            /*
             * Check if every setting is updated, then set is_profile_edit_completed to true
             * */

            if (Helper::isUserLocal()){
                if(
                    $user->bg_image &&
                    $user->profile_picture &&
                    $user->name &&
                    $user->about &&
                    $user->charges_per_hour &&
                    $user->currency &&
                    $user->language &&
                    $user->gender &&
                    UserLocation::where('user_id', $user->id)->get()->count() > 0 &&
                    UserActivity::where('user_id', $user->id)->get()->count() > 0
                ){
                    $user->is_profile_edit_completed = true;
                    $user->save();
                }
            }

            return response()->json([
                'user' => User::with('locations', 'locations.location', 'activities', 'activities.activity')->where('id', $user->id)->first(),
            ]);
        }catch (\Exception $e){
            return response()->json([
                'message' => $e->getMessage()
            ], 500);
        }

    }

    /**
     * @param Request $request
     * @return JsonResponse
     */
    public function get_user_general_settings(Request $request): JsonResponse
    {
        $user = $request->user();

        return response()->json([
            'user' => User::with('locations', 'locations.location', 'activities', 'activities.activity')->where('id', $user->id)->first(),
        ]);
    }

    /**
     * @param Request $request
     * @return JsonResponse
     */
    public function languages_get_all(Request $request): JsonResponse
    {
        return response()->json([
            'languages' => Helper::LANGUAGES,
        ]);
    }

    /**
     * @param Request $request
     * @return JsonResponse
     */
    public function currencies_get_all(Request $request): JsonResponse
    {
        return response()->json([
            'currencies' => Helper::CURRENCIES,
        ]);
    }
    /**
     * @param Request $request
     * @return JsonResponse
     */
    public function genders_get_all(Request $request): JsonResponse
    {
        return response()->json([
            'genders' => Helper::GENDERS,
        ]);
    }

    /**
     * @param Request $request
     * @return JsonResponse
     */
    public function update_user_role(Request $request): JsonResponse
    {
        $request->validate([
            'role' => 'required'
        ]);

        $user = $request->user();

        if ($request->role == Helper::LOCAL_TYPE){
            $user->roles()->detach();
            $user->assignRole(Helper::LOCAL_TYPE);
        }

        if ($request->role == Helper::TOURIST_TYPE){
            $user->roles()->detach();
            $user->assignRole(Helper::TOURIST_TYPE);
        }

        return response()->json([
            'message' => 'Role updated'
        ]);

    }

    /**
     * @param Request $request
     * @return JsonResponse
     */
    public function deleteLocal(Request $request): JsonResponse
    {
        $request->validate([
            'local_id' => 'required',
        ]);

        $user = $request->user();

        if ($user->hasRole(Helper::ADMIN_TYPE)){
            $local = User::find($request->local_id);
            if ($local){
                $local->delete();
                return response()->json([
                    'message' => 'Local Deleted'
                ]);
            }
            return response()->json([
                'message' => 'Local Not Found'
            ], 500);
        }
        return response()->json([
            'message' => 'Unauthorized.'
        ], 500);
    }

    public function sendRegistrationEmail($user)
    {
        $mailchimp = new ApiClient();
        $mailchimp->setApiKey("hpjqaafzFCHJPwcovvUczA");
        $message = "Thanks for registring on Show Around App. Your email verification code is: " . random_int(100000, 999999);
        $message_array = [
            'subject' => 'Welcome to Show Around.',
            'text' => $message,
            'from_email' => 'admin@maahey.com',
            'from_name' => 'Show Around',
            'to' => [
                [
                    'email' => $user->email,
                    'name' => $user->name,
                ]
            ],
        ];
        try{
            $mailchimp->messages->send(["message" => $message_array]);
            // $ex->getResponse()->getBody()->getContents()
        }catch(Exception $e){
            return false;
        }
    }

}
